/* OMMailMessage.m created by luke on Sat 18-Jul-1998 */


#import <OmniNetworking/OmniNetworking.h>
#import <OmniBase/OmniBase.h>
#import "OMMailMessage.h"

#define DEFAULT_SMTP_PORT 25

// SMTP Reply code strings

static NSString *serviceReadyGreetingCode = @"220";
static NSString *serviceClosingCode = @"221";
static NSString *actionOkayCode = @"250";
static NSString *startMailInputCode = @"354";

// Line termination character set

static NSCharacterSet *lineEndCharacterSet;

// Exception names

static NSString *OMSMTPException = @"OMSMTPException";


@interface OMMailMessage (PrivateAPI)

- (NSString *)mailPathForAddress:(NSString *)address;

- (NSArray *)rcptToCommandsForAddresses:(NSArray *)addresses;

- (void)confirmReply:(NSString *)reply matchedCode:(NSString *)smtpCode forCommandDescription:(NSString *)commandDescription;
    // Raises if reply doesn't match confirmation code.

@end

@implementation OMMailMessage

+ (void)initialize;
{
    static BOOL hasBeenInitialized = NO;

    if (hasBeenInitialized)
        return;

    [super initialize];

    lineEndCharacterSet = [NSCharacterSet characterSetWithCharactersInString:@"\r\n"];
}


- (id)init;
{
    return [self initWithReceiverAddresses:nil
                             senderAddress:nil
                               ccAddresses:nil
                              bccAddresses:nil
                                   subject:nil
                                      body:nil
                              smtpHostName:nil
                                  smtpPort:DEFAULT_SMTP_PORT];
}

- (id)initWithReceiverAddress:(NSString *)receiverAddress
                senderAddress:(NSString *)senderAddress
                      subject:(NSString *)subject
                         body:(NSString *)body
                 smtpHostName:(NSString *)hostName;
{
    return [self initWithReceiverAddresses:[NSArray arrayWithObject:receiverAddress]
                             senderAddress:senderAddress
                               ccAddresses:nil
                              bccAddresses:nil
                                   subject:subject
                                      body:body
                              smtpHostName:hostName
                                  smtpPort:DEFAULT_SMTP_PORT];
}

// Designated initializer
- (id)initWithReceiverAddresses:(NSArray *)receiverAddresses
                  senderAddress:(NSString *)senderAddress
                    ccAddresses:(NSArray *)ccAddresses
                   bccAddresses:(NSArray *)bccAddresses
                        subject:(NSString *)subject
                           body:(NSString *)body
                   smtpHostName:(NSString *)hostName
                       smtpPort:(unsigned int)port;
{
    if (!(self = [super init]))
        return nil;

    _receiverAddresses = [receiverAddresses mutableCopy];
    _senderAddress = [senderAddress retain];
    _ccAddresses = [ccAddresses mutableCopy];
    _bccAddresses = [bccAddresses mutableCopy];
    _subject = [subject retain];
    _body = [body retain];
    _smtpHostName = [hostName retain];
    _smtpPort = port;

    return self;
}

- (void)dealloc;
{
    [_receiverAddresses release];
    [_senderAddress release];
    [_ccAddresses release];
    [_bccAddresses release];
    [_subject release];
    [_body release];
    [_smtpHostName release];

    [super dealloc];
}

- (void)addReceiverAddress:(NSString *)receiverAddress;
{
    [_receiverAddresses addObject:receiverAddress];
}

- (NSMutableArray *)receiverAddresses;
{
    return _receiverAddresses;
}

- (void)setSenderAddress:(NSString *)senderAddress;
{
    if (_senderAddress != senderAddress) {
        [_senderAddress release];
        _senderAddress = [senderAddress retain];
    }
}

- (NSString *)senderAddress;
{
    return _senderAddress;
}

- (void)addCCAddress:(NSString *)ccAddress;
{
    [_ccAddresses addObject:ccAddress];
}

- (NSMutableArray *)ccAddresses;
{
    return _ccAddresses;
}

- (void)addBCCAddress:(NSString *)bccAddress;
{
    [_bccAddresses addObject:bccAddress];
}

- (NSMutableArray *)bccAddresses;
{
    return _bccAddresses;
}

- (void)setSubject:(NSString *)subject;
{
    if (_subject != subject) {
        [_subject release];
        _subject = [subject retain];
    }
}

- (NSString *)subject;
{
    return _subject;
}

- (void)setBody:(NSString *)body;
{
    if (_body != body) {
        [_body release];
        _body = [body retain];
    }
}

- (NSString *)body;
{
    return _body;
}

- (void)setSMTPHostName:(NSString *)hostName;
{
    if (_smtpHostName != hostName) {
        [_smtpHostName release];
        _smtpHostName = [hostName retain];
    }
}

- (NSString *)smtpHostName;
{
    return _smtpHostName;
}

- (void)setSMTPPort:(unsigned int)port;
{
    _smtpPort = port;
}

- (unsigned int )smtpPort;
{
    return _smtpPort;
}

- (void)deliver;
{
    NSString *heloCommand;
    NSString *mailFromCommand;
    NSMutableArray *rcptToCommands;
    NSString *smtpReply;
    NSString *dateLine;
    NSString *ccLine;
    NSCalendarDate *date;
    ONHost *smtpHost;
    ONTCPSocket *smtpSocket;
    unsigned int rcptToCommandIndex;
    unsigned int rcptToCommandCount;

    //
    //  Confirm that we have the minimum amount of information necessary to deliver a message.
    //

    if (![_senderAddress length]) {
        [NSException raise:NSInvalidArgumentException
                    format:@"Must specify sender address to deliver a message"];
    }
    
    if (![_receiverAddresses count] && [[_receiverAddresses objectAtIndex:0] length]) {
        [NSException raise:NSInvalidArgumentException
                    format:@"Must specify at least one receiver address to deliver a message"];
    }

    if (!(smtpHost = [ONHost hostForHostname:_smtpHostName])) {
        [NSException raise:NSGenericException
                    format:@"Unable to locate SMTP host: %@", _smtpHostName];
    }


    //
    //  Build helo command using local domain name.  OmniNetworking currently returns @"unknown.domain" if
    //  it failed to resolve the local domain name.  It would probably be better if it raised, instead.
    //

    heloCommand = [NSString stringWithFormat:@"helo %@\r\n", [ONHost domainName]];

    //
    //  Derive the 'mail from:' and 'rcpt to:' commands from the receiver and sender address(es).
    //

    mailFromCommand = [NSString stringWithFormat:@"mail from:%@\r\n", [self mailPathForAddress:_senderAddress]];
    rcptToCommands = [[[NSMutableArray alloc] init] autorelease];

    // Explicit receivers.
    [rcptToCommands addObjectsFromArray:[self rcptToCommandsForAddresses:_receiverAddresses]];

    // Cc receivers.
    [rcptToCommands addObjectsFromArray:[self rcptToCommandsForAddresses:_ccAddresses]];
    
    // Bcc receivers.
    [rcptToCommands addObjectsFromArray:[self rcptToCommandsForAddresses:_bccAddresses]];

    date = [NSCalendarDate date];
    dateLine = [NSString stringWithFormat:@"Date: %@\r\n",
        [date descriptionWithCalendarFormat:@"%a, %d %m %y %H:%M:%S %d %Z"]];

    smtpSocket = [ONTCPSocket tcpSocket];
    [smtpSocket connectToHost:smtpHost port:_smtpPort];

    NS_DURING {
        //
        //  Read and confirm service ready greeting.
        //

        smtpReply = [smtpSocket readString];
        [self confirmReply:smtpReply matchedCode:serviceReadyGreetingCode forCommandDescription:@"Initial connection"];

        //
        //  Initiate mail delivery process.  Send helo, confirm.  Send mail from command, confirm.
        //  send rcpt commands, confirm each one.  Send data, terminate data, confirm.  Quit.  Confirm.
        //

        [smtpSocket writeString:heloCommand];
        smtpReply = [smtpSocket readString];
        [self confirmReply:smtpReply matchedCode:actionOkayCode forCommandDescription:heloCommand];

        [smtpSocket writeString:mailFromCommand];
        smtpReply = [smtpSocket readString];
        [self confirmReply:smtpReply matchedCode:actionOkayCode forCommandDescription:mailFromCommand];

        for (rcptToCommandIndex = 0, rcptToCommandCount = [rcptToCommands count];
             rcptToCommandIndex < rcptToCommandCount; rcptToCommandIndex++) {
            NSString *rcptToCommand;

            rcptToCommand = [rcptToCommands objectAtIndex:rcptToCommandIndex];
            [smtpSocket writeString:rcptToCommand];
            smtpReply = [smtpSocket readString];
            [self confirmReply:smtpReply matchedCode:actionOkayCode forCommandDescription:rcptToCommand];
        }

        //
        //  Initial portion of data consists of headers.  Currently, Date:, From:, To:, Subject, Cc:.
        //  We don't write out Bcc: headers.
        //

        [smtpSocket writeString:@"data\r\n"];
        smtpReply = [smtpSocket readString];
        [self confirmReply:smtpReply matchedCode:startMailInputCode forCommandDescription:@"data\r\n"];

        [smtpSocket writeString:[NSString stringWithFormat:@"To: %@\r\n",
            [_receiverAddresses componentsJoinedByString:@",\r\n "]]];

        if ([_ccAddresses count] && [[_ccAddresses objectAtIndex:0] length]) {
            ccLine = [NSString stringWithFormat:@"Cc: %@\r\n", [_ccAddresses componentsJoinedByString:@",\r\n "]];
            [smtpSocket writeString:ccLine];
        }

        if ([_subject length])
            [smtpSocket writeString:[NSString stringWithFormat:@"Subject: %@\r\n", _subject]];

        [smtpSocket writeString:dateLine];
        [smtpSocket writeString:[NSString stringWithFormat:@"From: %@\r\n", [self mailPathForAddress:_senderAddress]]];

        // Write \r\n pair on empty line to indicate end of headers list
        [smtpSocket writeString:@"\r\n"];

        if ([_body length])
            [smtpSocket writeString:_body];


        //
        //  Terminate, confirm everything was received correctly, quit, confirm.
        //

        [smtpSocket writeString:@"\r\n.\r\n"];
        smtpReply = [smtpSocket readString];
        [self confirmReply:smtpReply matchedCode:actionOkayCode forCommandDescription:@"Data termination"];

        [smtpSocket writeString:@"quit\r\n"];
        smtpReply = [smtpSocket readString];
        [self confirmReply:smtpReply matchedCode:serviceClosingCode forCommandDescription:@"Quit"];
    } NS_HANDLER {
        [smtpSocket abortSocket];
        [localException raise]; // You're it.
    } NS_ENDHANDLER
    
    // Close SMTP socket, and we're done.
    [smtpSocket abortSocket];
}

@end

@implementation OMMailMessage (PrivateAPI)

//
//  Takes an email address of the form:
//
//      address@some.domain               or
//      <address@some.domain>             or
//      "Some Name" <address@some.domain>
//
//  and returns a mail path of the form:
//
//      <address@some.domain>
//

- (NSString *)mailPathForAddress:(NSString *)address;
{
    NSScanner *addressScanner;
    NSString *mailPath = nil;
    
    OBASSERT(address && [address length]);

    // Search for '<', which marks the beginning of the mail path.
    if (![address rangeOfString:@"<"].length) {
        // Doesn't contain '<', assume it's of form: address@some.domain
        
        return [NSString stringWithFormat:@"<%@>", address];
    }

    // Found '<', assume address is of form: xxx <address@some.domain>
    addressScanner = [[NSScanner alloc] initWithString:address];
    [addressScanner scanUpToString:@"<" intoString:(NSString **)nil];

    // Search for the trailing '>', which marks the end of the mail path.
    if (![[address substringFromIndex:[addressScanner scanLocation]] rangeOfString:@">"].length) {
        [addressScanner release];
        [NSException raise:NSGenericException
                    format:@"Unable to resolve mail path of form: <address@some.domain>, from address: %@", address];
    }
    
    [addressScanner scanUpToString:@">" intoString:&mailPath];
    [addressScanner release];

    // Apply trailing '>', which we scanned up to, but didn't read.
    return [mailPath stringByAppendingString:@">"];
}

//
//  Takes a list of addresses and returns the corresponding list of rcpt
//  to: commands, of the form:
//
//      rcpt to:<address@some.domain>
//
//  Addresses are translated to mail paths using -mailPathForAddress:
//

- (NSArray *)rcptToCommandsForAddresses:(NSArray *)addresses;
{
    NSMutableArray *rcptToCommands;
    unsigned int receiverIndex;
    unsigned int receiverCount;

    rcptToCommands = [[[NSMutableArray alloc] init] autorelease];
    
    for (receiverIndex = 0, receiverCount = [addresses count]; receiverIndex < receiverCount; receiverIndex++) {
        NSString *receiverAddress;
        NSString *rcptToCommand;

        receiverAddress = [addresses objectAtIndex:receiverIndex];
        rcptToCommand = [NSString stringWithFormat:@"rcpt to:%@\r\n", [self mailPathForAddress:receiverAddress]];
        [rcptToCommands addObject:rcptToCommand];
    }

    return rcptToCommands;
}

//
//  Raises if reply doesn't match confirmation code.  For example,
//
//      [self confirmReply:@"250 <luke>... Recipient ok" matchedCode:@"250" forCommandDescription:@"rcpt to:<luke>"]
//
//  ...would succeed.  Raises on failure.  NOTE: You should use well-defined
//  return code strings, rather than passing NSInlineCStrings.
//

- (void)confirmReply:(NSString *)reply matchedCode:(NSString *)smtpCode forCommandDescription:(NSString *)commandDescription;
{
    NSScanner *commandDescriptionScanner;
    NSString *oneLineCommandDescription;

    // Strip trailing linebreak from command description so the exception looks nicer.
    commandDescriptionScanner = [[[NSScanner alloc] init] autorelease];
    [commandDescriptionScanner scanUpToCharactersFromSet:lineEndCharacterSet intoString:&oneLineCommandDescription];
    
    if (![[reply substringToIndex:[smtpCode length]] isEqualToString:smtpCode]) {
        [NSException raise:OMSMTPException
                    format:@"Didn't receive expected reply code: %@, from command: %@.  Received: %@",
            smtpCode, oneLineCommandDescription, reply];
    }
}

@end

